/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.te;

import java.util.HashSet;
import java.util.Set;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import net.minecraft.block.Block;
import net.minecraft.block.state.IBlockState;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3i;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import net.dries007.tfc.TerraFirmaCraft;
import net.dries007.tfc.api.capability.IBellowsHandler;

import static net.minecraft.block.BlockHorizontal.FACING;

public class TEBellows extends TileEntity
{
    /**
     * The list of position offsets that bellows must check to input air into.
     * Directions go like this:
     * X: front of the bellows, positive values go forward
     * Y: vertical, self-explanatory, negative is below.
     * Z: X but rotated 90 degrees clockwise, positive values go right. in most cases you want this to be 0
     *
     * For example: a block that must sit right in front of the bellows(like fire pits) must register {@code Vec3i(1,0,0)},
     * {@link net.dries007.tfc.objects.blocks.BlockFirePit}
     * meanwhile, block that sink into ground(like forges) might want to register {@code Vec3i(1,-1,0)}.
     * {@link net.dries007.tfc.objects.blocks.BlockCharcoalForge}
     * If there is a guarantee for another block to register that position,
     * then there is no need do do it anymore with yours.
     */
    private static final Set<Vec3i> offsets = new HashSet<>();

    public static void addBellowsOffset(Vec3i offset)
    {
        offsets.add(offset);
    }

    private long lastPushed = 0L;

    // Min 0.125, Max 0.875
    @SideOnly(Side.CLIENT)
    public double getHeight()
    {
        int time = (int) (field_145850_b.func_82737_E() - lastPushed);
        if (time < 10)
            return (double) time * 0.075 + 0.125;
        else if (time < 20)
            return (double) (20 - time) * 0.075 + 0.125;
        return 0.125;
    }

    @Override
    public void func_145839_a(NBTTagCompound tag)
    {
        lastPushed = tag.func_74763_f("lastPushed");
        super.func_145839_a(tag);
    }

    @Override
    @Nonnull
    public NBTTagCompound func_189515_b(NBTTagCompound tag)
    {
        tag.func_74772_a("lastPushed", lastPushed);
        return super.func_189515_b(tag);
    }

    @Nullable
    @Override
    public SPacketUpdateTileEntity func_189518_D_()
    {
        return new SPacketUpdateTileEntity(field_174879_c, 127, func_189517_E_());
    }

    @Override
    @Nonnull
    public NBTTagCompound func_189517_E_()
    {
        return func_189515_b(new NBTTagCompound());
    }

    @Override
    public void onDataPacket(NetworkManager net, SPacketUpdateTileEntity pkt)
    {
        func_145839_a(pkt.func_148857_g());
        updateBlock();
    }

    public boolean onRightClick()
    {
        long time = field_145850_b.func_82737_E() - lastPushed;
        if (time < 20)
            return true;
        lastPushed = field_145850_b.func_82737_E();
        EnumFacing direction = field_145850_b.func_180495_p(field_174879_c).func_177229_b(field_185512_D); // It is a better idea to inherit the direction directly from the block.
        for (Vec3i offset : offsets)
        {
            BlockPos posx = field_174879_c.func_177981_b(offset.func_177956_o())
                .func_177967_a(direction, offset.func_177958_n())
                .func_177967_a(direction.func_176746_e(), offset.func_177952_p());
            Block block = field_145850_b.func_180495_p(posx).func_177230_c();
            if (block instanceof IBellowsHandler && ((IBellowsHandler) block).canIntakeFrom(this, offset, direction))
            {
                ((IBellowsHandler) block).onAirIntake(this, field_145850_b, posx, 1f);
                if (field_145850_b.field_72995_K)
                {
                    //TODO: actual sound, better particles and animation
                    // old bellows just shot particles outwards, that was simpler.
                    posx = field_174879_c.func_177972_a(direction);
                    field_145850_b.func_175688_a(EnumParticleTypes.SMOKE_LARGE, posx.func_177958_n() + .5d, posx.func_177956_o() + .5d, posx.func_177952_p() + .5d, 0, 0, 0);
                }
                return true;
            }
        }
        return false;
    }

    public void debug()
    {
        TerraFirmaCraft.getLog().debug("Debugging Bellows");
        TerraFirmaCraft.getLog().debug("Now: {} | Then: {} | Difference: {}", field_145850_b.func_82737_E(), lastPushed, (field_145850_b.func_82737_E() - lastPushed));
        TerraFirmaCraft.getLog().debug("Total Height: {}", getHeight());
    }

    private void updateBlock()
    {
        IBlockState state = field_145850_b.func_180495_p(field_174879_c);
        field_145850_b.func_184138_a(field_174879_c, state, state, 3); // sync TE
        func_70296_d(); // make sure everything saves to disk
    }
}
