/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.te;

import javax.annotation.Nonnull;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.block.state.IBlockState;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.ITickable;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.World;
import net.minecraftforge.fluids.FluidStack;
import net.minecraftforge.fluids.capability.CapabilityFluidHandler;
import net.minecraftforge.fluids.capability.IFluidHandler;

import net.dries007.tfc.ConfigTFC;
import net.dries007.tfc.TerraFirmaCraft;
import net.dries007.tfc.api.capability.heat.CapabilityItemHeat;
import net.dries007.tfc.api.capability.heat.IItemHeat;
import net.dries007.tfc.objects.blocks.BlockFirePit;
import net.dries007.tfc.objects.recipes.heat.HeatRecipe;
import net.dries007.tfc.objects.recipes.heat.HeatRecipeManager;
import net.dries007.tfc.util.Fuel;
import net.dries007.tfc.util.FuelManager;
import net.dries007.tfc.util.ITileFields;

import static net.dries007.tfc.api.capability.heat.CapabilityItemHeat.MAX_TEMPERATURE;
import static net.dries007.tfc.objects.blocks.BlockCharcoalForge.LIT;

@ParametersAreNonnullByDefault
public class TECharcoalForge extends TEInventory implements ITickable, ITileFields
{
    public static final int SLOT_FUEL_MIN = 0;
    public static final int SLOT_FUEL_MAX = 4;
    public static final int SLOT_INPUT_MIN = 5;
    public static final int SLOT_INPUT_MAX = 9;
    public static final int SLOT_EXTRA_MIN = 10;
    public static final int SLOT_EXTRA_MAX = 13;

    public static final int FIELD_TEMPERATURE = 0;

    private boolean requiresSlotUpdate = false;
    private float temperature; // Current Temperature
    private int burnTicks; // Ticks remaining on the current item of fuel
    private float burnTemperature; // Temperature provided from the current item of fuel
    private int airTicks; // Ticks of air provided by bellows

    public TECharcoalForge()
    {
        // 0 - 4 are fuel slots: 0 being the lowest, 4 highest (in order of consumption)
        // 5 - 9 are the input slots. Same arrangement (0 lowest, then alternating L, R, L high, R high)
        // 10 - 13 are the extra slots for molds and stuff. 10 at the top, 14 at the bottom. Iterate in that order
        super(14);

        temperature = 0;
        burnTemperature = 0;
        burnTicks = 0;
        airTicks = 0;
    }

    public void onAirIntake(float amount)
    {
        airTicks += (int) (200 * amount);
        if (airTicks > 600)
        {
            airTicks = 600;
        }
    }

    @Override
    public void func_73660_a()
    {
        if (field_145850_b.field_72995_K) return;
        IBlockState state = field_145850_b.func_180495_p(field_174879_c);
        if (state.func_177229_b(BlockFirePit.LIT))
        {
            // Update fuel
            if (burnTicks > 0)
            {
                // Double fuel consumption if using bellows
                burnTicks -= airTicks > 0 ? 2 : 1;
            }
            if (burnTicks <= 0)
            {
                // Consume fuel
                ItemStack stack = inventory.getStackInSlot(SLOT_FUEL_MIN);
                if (stack.func_190926_b())
                {
                    field_145850_b.func_175656_a(field_174879_c, state.func_177226_a(BlockFirePit.LIT, false));
                    burnTicks = 0;
                    burnTemperature = 0;
                }
                else
                {
                    inventory.setStackInSlot(SLOT_FUEL_MIN, ItemStack.field_190927_a);
                    Fuel fuel = FuelManager.getFuel(stack);
                    burnTicks += fuel.getAmount();
                    burnTemperature = fuel.getTemperature();
                }
            }
        }
        else if (burnTemperature > 0)
        {
            // If not lit, stop burning
            burnTemperature = 0;
            burnTicks = 0;
        }

        // Update bellows air
        if (airTicks > 0)
        {
            airTicks--;
        }
        else
        {
            airTicks = 0;
        }

        // Always update temperature / cooking, until the fire pit is not hot anymore
        if (temperature > 0 || burnTemperature > 0)
        {
            // Update temperature
            float targetTemperature = Math.min(MAX_TEMPERATURE, burnTemperature + airTicks);
            if (temperature < targetTemperature)
            {
                // Modifier for heating = 2x for bellows
                temperature += (airTicks > 0 ? 2 : 1) * ConfigTFC.GENERAL.temperatureModifierHeating;
            }
            else if (temperature > targetTemperature)
            {
                // Modifier for cooling = 0.5x for bellows
                temperature -= (airTicks > 0 ? 0.5 : 1) * ConfigTFC.GENERAL.temperatureModifierHeating;
            }

            // Update items in slots
            // Loop through input + 2 output slots
            for (int i = SLOT_INPUT_MIN; i <= SLOT_EXTRA_MAX; i++)
            {
                ItemStack stack = inventory.getStackInSlot(i);
                IItemHeat cap = stack.getCapability(CapabilityItemHeat.ITEM_HEAT_CAPABILITY, null);
                if (cap != null)
                {
                    // Update temperature of item
                    float itemTemp = cap.getTemperature();
                    if (temperature > itemTemp)
                    {
                        CapabilityItemHeat.addTemp(cap, (float) ConfigTFC.GENERAL.temperatureModifierItemHeating);
                    }

                    // This will melt + consume the input stack
                    // Output stacks are assumed to not melt (see the case of ceramic molds in the output)
                    if (cap.isMolten() && i <= SLOT_INPUT_MAX)
                    {
                        handleInputMelting(stack, i);
                    }
                }
            }
        }

        // This is here to avoid duplication glitches
        if (requiresSlotUpdate)
        {
            cascadeFuelSlots();
        }
    }

    public void onCreate()
    {
        burnTicks += 200;
        burnTemperature = 500;
    }

    @Override
    public void setAndUpdateSlots(int slot)
    {
        this.func_70296_d();
        requiresSlotUpdate = true;
    }

    @Override
    public int getSlotLimit(int slot)
    {
        // All slots have limit 1
        return 1;
    }

    @Override
    public boolean isItemValid(int slot, ItemStack stack)
    {
        if (slot <= SLOT_FUEL_MAX)
        {
            // Fuel slots - anything that is a valid TFC fuel
            return FuelManager.isItemForgeFuel(stack);
        }
        else if (slot <= SLOT_INPUT_MAX)
        {
            // Input slots - anything that can heat up
            return stack.hasCapability(CapabilityItemHeat.ITEM_HEAT_CAPABILITY, null);
        }
        else
        {
            // Extra slots - anything that can heat up and hold fluids
            return stack.hasCapability(CapabilityFluidHandler.FLUID_HANDLER_CAPABILITY, null) && stack.hasCapability(CapabilityItemHeat.ITEM_HEAT_CAPABILITY, null);
        }
    }

    @Override
    public void func_145839_a(NBTTagCompound nbt)
    {
        temperature = nbt.func_74760_g("temperature");
        burnTicks = nbt.func_74762_e("burnTicks");
        airTicks = nbt.func_74762_e("airTicks");
        burnTemperature = nbt.func_74760_g("burnTemperature");
        super.func_145839_a(nbt);
    }

    @Override
    @Nonnull
    public NBTTagCompound func_189515_b(NBTTagCompound nbt)
    {
        nbt.func_74776_a("temperature", temperature);
        nbt.func_74768_a("burnTicks", burnTicks);
        nbt.func_74768_a("airTicks", airTicks);
        nbt.func_74776_a("burnTemperature", burnTemperature);
        return super.func_189515_b(nbt);
    }

    public void debug()
    {
        TerraFirmaCraft.getLog().debug("Debugging Charcoal Forge:");
        TerraFirmaCraft.getLog().debug("Temp {} | Burn Temp {} | Fuel Ticks {}", temperature, burnTemperature, burnTicks);
        TerraFirmaCraft.getLog().debug("Burning? {}", field_145850_b.func_180495_p(field_174879_c).func_177229_b(LIT));
        for (int i = SLOT_INPUT_MIN; i <= SLOT_INPUT_MAX; i++)
        {
            TerraFirmaCraft.getLog().debug("Slot: {} - NBT: {}", i, inventory.getStackInSlot(i).serializeNBT().toString());
        }
    }

    @Override
    public boolean shouldRefresh(World world, BlockPos pos, IBlockState oldState, IBlockState newSate)
    {
        return oldState.func_177230_c() != newSate.func_177230_c();
    }

    private void handleInputMelting(ItemStack stack, int startIndex)
    {
        HeatRecipe recipe = HeatRecipeManager.get(stack);
        IItemHeat cap = stack.getCapability(CapabilityItemHeat.ITEM_HEAT_CAPABILITY, null);

        if (recipe != null && cap != null)
        {
            // Handle possible metal output
            FluidStack fluidStack = recipe.getOutputMetal(stack);
            float itemTemperature = cap.getTemperature();
            if (fluidStack != null)
            {
                // Loop through all input slots
                for (int i = SLOT_EXTRA_MIN; i <= SLOT_EXTRA_MAX; i++)
                {
                    // While the fluid is still waiting
                    if (fluidStack.amount <= 0)
                    {
                        break;
                    }
                    // Try an output slot
                    ItemStack output = inventory.getStackInSlot(i);
                    // Fill the fluid
                    IFluidHandler fluidHandler = output.getCapability(CapabilityFluidHandler.FLUID_HANDLER_CAPABILITY, null);
                    if (fluidHandler != null)
                    {
                        int amountFilled = fluidHandler.fill(fluidStack.copy(), true);
                        if (amountFilled > 0)
                        {
                            fluidStack.amount -= amountFilled;

                            // If the fluid was filled, make sure to make it the same temperature
                            IItemHeat heatHandler = output.getCapability(CapabilityItemHeat.ITEM_HEAT_CAPABILITY, null);
                            if (heatHandler != null)
                            {
                                heatHandler.setTemperature(itemTemperature);
                            }
                        }
                    }
                }
            }

            // Handle possible item output
            ItemStack outputStack = recipe.getOutputStack();
            if (outputStack != null && !outputStack.func_190926_b())
            {
                // Loop through all input slots
                for (int i = SLOT_EXTRA_MAX; i <= SLOT_EXTRA_MAX; i++)
                {
                    outputStack = inventory.insertItem(i, outputStack, false);
                    if (!outputStack.func_190926_b())
                    {
                        inventory.insertItem(i, outputStack, false);
                    }
                }
            }

            // Handle removal of input
            ItemStack inputStack = recipe.consumeInput(stack);
            inventory.setStackInSlot(startIndex, inputStack);
        }
    }

    private void cascadeFuelSlots()
    {
        // This will cascade all fuel down to the lowest available slot
        int lowestAvailSlot = 0;
        for (int i = 0; i <= SLOT_FUEL_MAX; i++)
        {
            ItemStack stack = inventory.getStackInSlot(i);
            if (!stack.func_190926_b())
            {
                // Move to lowest avail slot
                if (i > lowestAvailSlot)
                {
                    inventory.setStackInSlot(lowestAvailSlot, stack.func_77946_l());
                    inventory.setStackInSlot(i, ItemStack.field_190927_a);
                }
                lowestAvailSlot++;
            }
        }
        requiresSlotUpdate = false;
    }

    @Override
    public int getFieldCount()
    {
        return 1;
    }

    @Override
    public void setField(int index, int value)
    {
        if (index == FIELD_TEMPERATURE)
        {
            this.temperature = (float) value;
        }
        else
        {
            TerraFirmaCraft.getLog().warn("Invalid field ID {} in TECharcoalForge#setField", index);
        }
    }

    @Override
    public int getField(int index)
    {
        if (index == FIELD_TEMPERATURE)
        {
            return (int) temperature;
        }
        TerraFirmaCraft.getLog().warn("Invalid field ID {} in TECharcoalForge#getField", index);
        return 0;
    }
}
